@@ -10,6 +10,7 @@
            #:def-c-type #:def-c-enum #:def-c-struct #:def-c-var #:def-call-out))
 
 (ffi:default-foreign-language :stdc-stdcall) ; WINAPI means __stdcall
+(eval-when (compile) (setq ffi:*foreign-guard* t))
 
 (in-package "W32")
 
@@ -48,9 +49,45 @@
 (def-call-out GetCurrentProcessId (:library kernel32)
   (:arguments) (:return-type dword))
 
+(def-call-out GetProcessVersion (:library kernel32)
+  (:arguments (pid dword)) (:return-type dword))
+
+;;; longhorn
+(def-call-out GetProcessId (:library kernel32)
+  (:arguments (hProcess handle)) (:return-type dword))
+
+(def-call-out GetCurrentThreadId (:library kernel32)
+  (:arguments) (:return-type dword))
+
+;;; longhorn
+(def-call-out GetProcessIdOfThread (:library kernel32)
+  (:arguments (hThread handle)) (:return-type dword))
+
+;;; longhorn
+(def-call-out GetThreadId (:library kernel32)
+  (:arguments (hThread handle)) (:return-type dword))
+
+(def-call-out GetThreadPriority (:library kernel32)
+  (:arguments (hThread handle)) (:return-type int))
+
+(def-call-out GetThreadPriorityBoost (:library kernel32)
+  (:arguments (hThread handle) (no-boost (c-ptr boolean) :out))
+  (:return-type boolean))
+
+(def-call-out GetProcessPriorityBoost (:library kernel32)
+  (:arguments (hProcess handle) (no-boost (c-ptr boolean) :out))
+  (:return-type boolean))
+
 (def-call-out CloseHandle (:library kernel32)
   (:arguments (handle handle)) (:return-type boolean))
 
+(defmacro with-handle ((handle form) &body forms)
+  (multiple-value-bind (body-rest declarations) (SYSTEM::PARSE-BODY forms)
+    `(let ((,handle ,form))
+       (declare (read-only ,handle) ,@declarations)
+       (unwind-protect (progn ,@body-rest)
+         (when ,handle (CloseHandle ,handle))))))
+
 ;; (c-lines "#include <winnt.h>~%")
 (eval-when (compile eval load)
 (def-c-enum RIGHTS
@@ -136,6 +173,66 @@
  (win32:DestroyIcon icon)
 |#
 
+;; buttons
+(defconstant MB_OK 0)
+(defconstant MB_OKCANCEL 1)
+(defconstant MB_ABORTRETRYIGNORE 2)
+(defconstant MB_YESNOCANCEL 3)
+(defconstant MB_YESNO 4)
+(defconstant MB_RETRYCANCEL 5)
+(defconstant MB_CANCELTRYCONTINUE 6)
+;; f1?
+(defconstant MB_HELP #x4000)
+;; icons
+(defconstant MB_ICONEXCLAMATION #x30)
+(defconstant MB_ICONWARNING #x30)
+(defconstant MB_ICONINFORMATION 64)
+(defconstant MB_ICONASTERISK 64)
+(defconstant MB_ICONQUESTION 32)
+(defconstant MB_ICONSTOP 16)
+(defconstant MB_ICONERROR 16)
+(defconstant MB_ICONHAND 16)
+;; defaults
+(defconstant MB_DEFBUTTON1 0)
+(defconstant MB_DEFBUTTON2 256)
+(defconstant MB_DEFBUTTON3 512)
+(defconstant MB_DEFBUTTON4 #x300)
+;; modality
+(defconstant MB_APPLMODAL 0)
+(defconstant MB_SYSTEMMODAL 4096)
+(defconstant MB_TASKMODAL #x2000)
+
+;; desktop
+(defconstant MB_DEFAULT_DESKTOP_ONLY #x20000)
+
+;; text
+(defconstant MB_RIGHT #x80000)
+(defconstant MB_RTLREADING #x100000)
+(defconstant MB_SETFOREGROUND #x10000)
+(defconstant MB_TOPMOST #x40000)
+;; depends on windows version MB_SERVICE_NOTIFICATION
+;; MB_SERVICE_NOTIFICATION_NT3X
+
+(defconstant IDABORT 3 "Abort button was selected.")
+(defconstant IDCANCEL 2 "Cancel button was selected.")
+(defconstant IDCONTINUE 11 "Continue button was selected.")
+(defconstant IDIGNORE 5 "Ignore button was selected.")
+(defconstant IDNO 7 "No button was selected.")
+(defconstant IDOK 1 "OK button was selected.")
+(defconstant IDRETRY 4 "Retry button was selected.")
+(defconstant IDTRYAGAIN 10 "Try Again button was selected.")
+(defconstant IDYES 6 "Yes button was selected.")
+
+(def-call-out MessageBoxA (:library user32)
+  (:arguments (parent handle) (text c-string) (caption c-string) (type uint))
+  (:documentation "http://msdn.microsoft.com/library/en-us/winui/winui/windowsuserinterface/windowing/dialogboxes/dialogboxreference/dialogboxfunctions/messagebox.asp")
+  (:return-type int))
+#| examples:
+ (win32:MessageBoxA nil "welcome to clisp" "message from clisp" win32:MB_OK)
+ (win32:MessageBoxA nil "welcome to clisp" "message from clisp"
+                    (logior win32:MB_YESNOCANCEL win32:MB_ICONWARNING))
+|#
+
 (eval-when (compile eval load)
   (defconstant BUFSIZ 4096)     ; <stdio.h>
   (defconstant MAX_PATH 260))   ; <windef.h>
@@ -161,6 +258,22 @@
   (:arguments (title c-string))
   (:return-type boolean))
 
+(def-call-out GetConsoleWindow (:library kernel32)
+  (:documentation "http://msdn.microsoft.com/library/en-us/dllproc/base/getconsolescreenbufferinfo.asp")
+  (:arguments) (:return-type handle))
+
+(def-c-struct SMALL_RECT (Left short) (Top short) (Right short) (Bottom short))
+(def-c-struct COORD (X short) (Y short))
+(def-c-struct CONSOLE_SCREEN_BUFFER_INFO
+  (dwSize COORD) (dwCursorPosition COORD)
+  (wAttributes word) (srWindow SMALL_RECT) (dwMaximumWindowSize COORD))
+
+(def-call-out GetConsoleScreenBufferInfo (:library kernel32)
+  (:documentation "http://msdn.microsoft.com/library/en-us/dllproc/base/getconsolescreenbufferinfo.asp")
+  (:arguments (StdHandle handle)
+              (csbiInfo (c-ptr CONSOLE_SCREEN_BUFFER_INFO) :out :alloca))
+  (:return-type boolean))
+
 ;; system information
 (def-call-out GetSystemDirectoryA (:library kernel32)
   (:arguments (buffer (c-ptr (c-array-max character #.MAX_PATH)) :out :alloca)
@@ -328,23 +441,7 @@ Initially, this is a handle to the active console screen buffer, CONOUT$.")
   (:return-type boolean))
 
 
-#|
-
- (defun check-all (enum-type function buf-size)
-  (format t "~&;; ~s:~%" function)
-  (maphash (lambda (key val)
-             (let ((res (multiple-value-list (funcall function key buf-size))))
-               (format t " ~S -> ~S~@[ ~S~]~%" val res
-                       (unless (car res) (w32:GetLastError)))))
-           (ffi::enum-table enum-type)))
- (check-all 'w32:EXTENDED_NAME_FORMAT 'w32:GetUserNameExA w32::UNLEN)
- (check-all 'w32:EXTENDED_NAME_FORMAT 'w32:GetComputerObjectNameA w32::UNLEN)
- (check-all 'w32:COMPUTER_NAME_FORMAT 'w32:GetComputerNameExA
-            w32::MAX_COMPUTERNAME_LENGTH)
-
-|#
-
-;;; ==========================================================================
+;; ==========================================================================
 
 (pushnew "WIN32" custom:*system-package-list* :test #'string=)
 (provide "win32")
