@@ -2,12 +2,11 @@
 ; Examples to test the Oracle CLISP library This is VERY incomplete,
 ; but should at least tell you if you built it right.
 
-; $Id: examples.lisp,v 1.1 2002/09/11 14:27:47 hin Exp $
+; $Id: examples.lisp,v 1.2 2002/09/20 13:51:54 hin Exp $
 
 ; Replace the connect info with your user and server
 
-; (oracle:connect "myuser" "mypasswd" "myserver")
-(oracle:connect "alma" "alma" "")
+(oracle:connect "myuser" "mypasswd" "myserver")
 
 ; Create a table and insert some data
 (oracle:run-sql "CREATE TABLE foo (i number(7), j number(7))")
