@@ -17,9 +17,9 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
-   $Id: orafns.c,v 1.1 2002/09/11 14:27:58 hin Exp $
+   $Id: orafns.c,v 1.17 2005/08/26 13:47:23 hin Exp $
 
 */
 
@@ -37,12 +37,18 @@
 /* Constants */
 
 /* Default no. of bytes for row prefetch */
-#define DEFAULT_PREFETCH_BYTES  1024 * 1024
+#define DEFAULT_PREFETCH_BYTES  65536
 /* Size of per-connection error buffer */
 #define ERRBUF_BYTES            100000
+/* Truncate echoed parameter values to this length */
+#define ERRBUF_PARAM_TRUNC      500
+/* Default size of LONG fetch */
+#define DEFAULT_LONG_LEN		500000
+
+#define minof(a, b) (a < b ? a : b)
 
 /* Local routines, w/ wrappers passed to external interface */
-static struct db_conn * connect(char *, char *, char *, char *, int, int);
+static struct db_conn * connect(char *, char *, char *, char *, int, int, int, int);
 static int              disconnect(struct db_conn *);
 static int              exec_sql(struct db_conn *, char *, struct sqlparam **, int);
 static int              fetch_row(struct db_conn *);
@@ -57,12 +63,14 @@ static int              rollback(struct db_conn *);
 static int              set_auto_commit(struct db_conn *, int);
 
 /* Purely local routines */
-static void      append_oci_error(char *, OCIError *);
+static void      append_oci_error(char *, void *, int);
+static void      append_indicator(struct db_conn *, char *);
 static int       init_session(struct db_conn *, char *, char *, char *, char*, int);
 static int       get_cols(struct db_conn *);
 static int       get_param_attr(CONST dvoid *, int, struct db_conn *, dvoid *, ub4 *, char *, ub4);
 static char *    decode_data_type(int);
-static int       fetch_data_len(int, int);
+static int       fetch_data_len(int, int, int);
+static int       is_blob_type(int);
 static void      free_columns(struct db_conn *);
 static void      free_column(struct column *);
 static void      free_if_non_null(void *);
@@ -71,6 +79,7 @@ static int       check_active_statement(struct db_conn *, char *);
 static int       check_active_select(struct db_conn *, char *);
 static int       empty(char *);
 static char *    valid_string(char *);
+static void      hexify (unsigned char *, int) ;
 
 /* Static indicator variables for input data */
 static sb4      null_indicator = -1;
@@ -102,8 +111,11 @@ int main(int argc, char **argv) {
 
   /* Connect to server */
   db = connect(argv[1], argv[2], argv[3], argv[4],
-                      -1,   /* Prefetch bytes */
-                      1);   /* Auto-commit? */
+			   -1,   /* Prefetch bytes */
+			   1,    /* Auto-commit? */
+			   0,    /* Long read len */
+			   0);   /* Truncate OK? */
+			          
   if ( ! db ) {
     /* Should never really get this */
     printf("Error initting connection\n");
@@ -175,7 +187,7 @@ int main(int argc, char **argv) {
 
 /* Init a new connection, given user, schema, password and SID */
 
-static struct db_conn * connect(char * user, char * schema, char * password, char * sid, int prefetch_bytes, int auto_commit)
+static struct db_conn * connect(char * user, char * schema, char * password, char * sid, int prefetch_bytes, int auto_commit, int long_len, int truncate_ok)
 {
   sb4       status  = OCI_SUCCESS;
 
@@ -197,9 +209,11 @@ static struct db_conn * connect(char * user, char * schema, char * password, cha
   if ( ! db->success )
     return db;
 
-  /* Copy buffer size and auto-commit */
+  /* Other params: fetch buffer size, auto-commit, LONG fetch size, truncate-allowed */
   db->prefetch_bytes = prefetch_bytes;
   db->auto_commit = auto_commit;
+  db->long_len = long_len >= 0 ? long_len : DEFAULT_LONG_LEN;
+  db->truncate_ok = truncate_ok;
 
   /* Success */
   db->success = 1;
@@ -273,7 +287,7 @@ static int fetch_row(struct db_conn * db)
   if ( fetch_status != OCI_SUCCESS && fetch_status != OCI_SUCCESS_WITH_INFO ) {
     /* Error fetching */
     sprintf(db->errmsg, "Fetch error %d:\n", fetch_status);
-    append_oci_error(db->errmsg, db->err);
+    append_oci_error(db->errmsg, db->err, 0);
     return db->success = 0;
   }
   
@@ -291,7 +305,7 @@ static int fetch_row(struct db_conn * db)
       sprintf(db->errmsg, "Error getting error status checking for EOF\n");
       return db->success = 0;
     }
-    
+
     /* Check for EOF */
     if ( errcode == 1403 ) {
       db->eof = 1;
@@ -308,24 +322,102 @@ static int fetch_row(struct db_conn * db)
       truncated = 1;
       null = 1;
     }
+	/* Ignored warnings */
+	else if (    errcode == 24347    /* NULL column in aggregate function */
+			  || errcode == 24348 )  /* DELETE or UPDATE without WHERE */
+	  ; /* Ignore */
     else {
       sprintf(db->errmsg, "Unknown Oracle warning %d\n", errcode);
-      append_oci_error(db->errmsg, db->err);
+      append_oci_error(db->errmsg, db->err, 0);
       return db->success = 0;
     }
   }
+
+  /* If row-level truncation indicator given, make sure it was not because of
+	 a ROWID being in the result (which will always trigger that indicator).
+	 In this case, scan the result columns for a ROWID type, and if one appears,
+	 turn off the global truncation and NULL flags */
+  if ( truncated )
+	for ( n=0; n < db->ncol; n++ ) {
+	  col = &db->columns[n];
+	  if ( col->dtype == SQLT_RDD ) {
+		truncated = 0;
+		null = 0;
+		/* Platform-dependent ROWID Descriptor gives its length in the indicator, usually 18 bytes */
+		col->indicator = 0; /* Reset it as if not truncated */
+	  }
+	}
+	  
+  /* Positive indicator means truncated because it was that large, -2 means truncated from value so large, size does not fit in a SB2 */
+  for ( n=0; n < db->ncol; n++ ) {
+    col = &db->columns[n];
+	if ( col->indicator > 0 || col->indicator == -2 )
+	  truncated = 1;
+  }
+
+  /* If data was truncated and that is not allowed, raise an error */
+  if ( truncated && ! db->truncate_ok ) {
+	char * defind = db->long_len == DEFAULT_LONG_LEN ? "default" : "specified";
+	sprintf(db->errmsg,
+			"Size of a LONG column exceeds %s maximum of %d.\nYou must either increase size of LONG/BLOB fetch buffer, or set flag allowing truncation.\n", defind, db->long_len);
+	append_oci_error(db->errmsg, db->err, 0);
+	return db->success = 0;
+  }
   
-  /* Copy is-null indicator status to external structures.  */
+  /* Loop over columns, fetching propagating is-null status and fetching BLOB data */
   for ( n=0; n < db->ncol; n++ ) {
     struct sqlval * r = db->currow[n];
     col = &db->columns[n];
-    if ( col->indicator == 0 )
-      r->is_null = 0;
+
+	if ( col->indicator == 0 ) {
+
+		r->is_null = 0;
+
+		/* Fetch BLOB data using locator, checking it fits */
+		if ( is_blob_type(col->dtype) ) {
+		  sword lob_read_status = 0;
+		  OCILobLocator * lob = col->lob_locator;
+		  ub4 try_nread = db->long_len + 1;
+		  ub4 act_nread = try_nread;
+		  if ( ! lob ) {
+			sprintf(db->errmsg, "Internal error: no LobLocator for column %d\n", n);
+			return db->success = 0;
+		  }
+		  lob_read_status = OCILobRead(db->svc, db->err, lob, &act_nread, 1,
+									   col->data, try_nread, 0, 0, 0, SQLCS_IMPLICIT);
+
+		  /* If could not fit the whole thing in the buffer, will have read one beyond db->long_len,
+			 or perhaps Oracle will return OCI_NEED_DATA?  If that is the case, see if we
+			 allow truncating it.  */
+		  if ( ( act_nread >= try_nread || lob_read_status == OCI_NEED_DATA ) && ! db->truncate_ok ) {
+			char * defind = db->long_len == DEFAULT_LONG_LEN ? "default" : "specified";
+			sprintf(db->errmsg,
+					"Size of BLOB/CLOB column '%s' exceeds %s maximum of %d.\n"
+					"You must either increase size of LONG/BLOB fetch buffer, or set flag allowing truncation.\n",
+					col->name, defind, db->long_len);
+			append_oci_error(db->errmsg, db->err, 0);
+			return db->success = 0;
+		  }
+		  else if ( lob_read_status != OCI_SUCCESS ) {
+			sprintf(db->errmsg, "Error reading LOB column '%s'\n", col->name);
+			append_oci_error(db->errmsg, db->err, 0);
+			return db->success = 0;
+		  }
+		
+                  /* For BLOB or BFILE, convert to hex */
+                  if ( col->dtype == SQLT_BLOB || col->dtype == SQLT_BFILE )
+                    hexify((unsigned char *) col->data, act_nread);
+                  else
+                    /* Null-terminate the data buffer */
+                    ((char *) col->data)[minof(act_nread, db->long_len)] = '\0';
+		}
+	}
     else if ( col->indicator == -1 )
       r->is_null = 1;
-    else {
+    else if ( col->indicator < -2 ) {
       sprintf(db->errmsg, "Bad indicator value %d for column %d\n", col->indicator, n);
-      append_oci_error(db->errmsg, db->err);
+      append_oci_error(db->errmsg, db->err, 0);
+	  return db->success = 0;
     }
   }
   
@@ -355,6 +447,9 @@ static int get_cols(struct db_conn * db)
   int               fetch_buflen        = 0;
   int               success             = 0;
   sb4               define_status       = OCI_SUCCESS;
+  dvoid *		    define_dest        = 0;
+  sb4               define_len          = 0;
+  ub2               define_type         = 0;
 
   /* Clear out previous results */
   free_columns(db);
@@ -368,7 +463,7 @@ static int get_cols(struct db_conn * db)
   if ( attr_status != OCI_SUCCESS ) {
     /* Error getting param info */
     sprintf(db->errmsg, "Error getting param count:\n");
-    append_oci_error(db->errmsg, db->err);
+    append_oci_error(db->errmsg, db->err, 0);
     return db->success = 0;
   }
 
@@ -381,6 +476,10 @@ static int get_cols(struct db_conn * db)
 
   do {
 
+	/* Oracle requires a sb2, not an ub1, to receive the OCI_ATTR_PRECISION for implicit describes,
+	   even though in the data dictionary it is a one-byte unsigned value.  Got that?  I didn't.  */
+	sb2 tmp_precision = 0;
+
     ncol++;
     param_status = OCIParamGet(db->stmt, OCI_HTYPE_STMT, db->err, (dvoid **) &param, ncol);
 
@@ -415,7 +514,7 @@ static int get_cols(struct db_conn * db)
       else {
         /* A real error */
         sprintf(db->errmsg, "Error getting param col %d:\n", ncol);
-        append_oci_error(db->errmsg, db->err);
+        append_oci_error(db->errmsg, db->err, 0);
         return db->success = 0;
       }
     }
@@ -432,11 +531,12 @@ static int get_cols(struct db_conn * db)
     col = &db->columns[ncol-1];
 
     /* Get param attributes */
+	/* Note for OCI_ATTR_PRECISION we receive the value in an (sb2 *) rather than (ub1 *) for our "implicit describe" */
     success = 1;
     success = success && get_param_attr(param, ncol, db, &colnamep,       &colname_len, "name",      OCI_ATTR_NAME);
     success = success && get_param_attr(param, ncol, db, &col->dtype,     0,            "data type", OCI_ATTR_DATA_TYPE);
     success = success && get_param_attr(param, ncol, db, &col->dsize,     0,            "size",      OCI_ATTR_DATA_SIZE);
-    success = success && get_param_attr(param, ncol, db, &col->precision, 0,            "precision", OCI_ATTR_PRECISION);
+    success = success && get_param_attr(param, ncol, db, &tmp_precision,  0,            "precision", OCI_ATTR_PRECISION);
     success = success && get_param_attr(param, ncol, db, &col->scale,     0,            "scale",     OCI_ATTR_SCALE);
     success = success && get_param_attr(param, ncol, db, &col->null_ok,   0,            "is-null",   OCI_ATTR_IS_NULL);
     if ( ! success )
@@ -447,6 +547,13 @@ static int get_cols(struct db_conn * db)
     colname[colname_len] = '\0';
     col->name = (char *) strdup(colname);
 
+	/* Copy precision.  For grins, check that it will fit in an ub1 */
+	if ( tmp_precision < 0 || tmp_precision > 127 ) {
+	  sprintf(db->errmsg, "Precision [%d] out of bounds for column '%s'", col->name);
+	  return db->success = 0;
+	}
+	col->precision = tmp_precision;
+	
     /*
       printf("Col %2d: type is %-10s name = %-20s size = %4d precision = %2d scale = %2d null-ok=%d\n",
              ncol, decode_data_type(col->dtype), col->name, col->dsize, col->precision, col->scale, col->null_ok);
@@ -455,17 +562,39 @@ static int get_cols(struct db_conn * db)
     /* Set up a define variable for this col to get its fetched value.
        Note that we ask Oracle to convert everything to
        null-terminated string, SQLT_STR. */
-    fetch_buflen = fetch_data_len(col->dtype, col->dsize);
+    fetch_buflen = fetch_data_len(col->dtype, col->dsize, db->long_len);
     col->data = malloc(fetch_buflen);
     *((char *) col->data) = '\0';
     col->indicator = 0;
     col->nfetched = 0;
     col->rcode = 0;
-    define_status = OCIDefineByPos(db->stmt, &col->def, db->err, ncol, col->data, fetch_buflen, SQLT_STR,
+	col->lob_locator = 0;
+	
+	/* Fetch LOB's as LobLocator's, everything else as SQLT_STR */
+	if ( is_blob_type(col->dtype) ) {
+	  /* Allocate LOB locator */
+	  sword dalloc_stat = OCIDescriptorAlloc(db->env, (void *) &col->lob_locator, OCI_DTYPE_LOB, 0, 0);
+	  if ( dalloc_stat != OCI_SUCCESS ) {
+		sprintf(db->errmsg, "Error creating LOB descriptor for column '%s'", col->name);
+		return db->success = 0;
+	  }
+
+	  define_dest = &col->lob_locator;
+	  define_len = 0;
+	  define_type = col->dtype;
+	}
+	else {
+	  define_dest = col->data;
+	  define_len = fetch_buflen;
+	  define_type = SQLT_STR;
+	}
+	
+	/* Do the Define */
+    define_status = OCIDefineByPos(db->stmt, &col->def, db->err, ncol, define_dest, define_len, define_type,
                                    &col->indicator, &col->nfetched, &col->rcode, OCI_DEFAULT);
     if ( define_status != OCI_SUCCESS ) {
       sprintf(db->errmsg, "Error setting up define for column '%s':\n", col->name);
-      append_oci_error(db->errmsg, db->err);
+      append_oci_error(db->errmsg, db->err, 0);
       return db->success = 0;
     }
   }
@@ -515,7 +644,7 @@ static int commit(struct db_conn * db)
   status = OCITransCommit(db->svc, db->err, OCI_DEFAULT);
   if ( status != OCI_SUCCESS ) {
     sprintf(db->errmsg, "Error commiting transaction:\n");
-    append_oci_error(db->errmsg, db->err);
+    append_oci_error(db->errmsg, db->err, 0);
     return db->success = 0;
   }
 
@@ -542,7 +671,7 @@ static int rollback(struct db_conn * db)
   status = OCITransRollback(db->svc, db->err, OCI_DEFAULT);
   if ( status != OCI_SUCCESS ) {
     sprintf(db->errmsg, "Error rolling back transaction:\n");
-    append_oci_error(db->errmsg, db->err);
+    append_oci_error(db->errmsg, db->err, 0);
     return db->success = 0;
   }
 
@@ -594,20 +723,26 @@ static int exec_sql(struct db_conn * db, char * sql, struct sqlparam ** params,
      active transaction currently in progress with the old handle?  It
      probably should.  */
   if ( db->stmt ) {
+	/* Note - freeing the statement handle recursively frees its
+	   children "define" handles, and since we may have kept around
+	   copies of those handles, we need to clear out our copies
+	   first. */
+	free_columns(db);
     OCIHandleFree(db->stmt, OCI_HTYPE_STMT);
     db->stmt = 0;
   }
   if ( OCI_SUCCESS !=  OCIHandleAlloc(db->env, (dvoid **) &db->stmt, OCI_HTYPE_STMT, 0, 0) ) {
     sprintf(db->errmsg, "Error allocating statement handle\n");
-    append_oci_error(db->errmsg, db->err);
+    append_oci_error(db->errmsg, db->err, 0);
     return db->success = 0;
   }
 
   /* "Prepare" the statement - this rarely returns an error */
-  status = OCIStmtPrepare(db->stmt, db->err, sql, strlen(sql), OCI_V8_SYNTAX, OCI_DEFAULT);
+  status = OCIStmtPrepare(db->stmt, db->err, sql, strlen(sql), OCI_NTV_SYNTAX, OCI_DEFAULT);
   if ( status != OCI_SUCCESS ) {
-    sprintf(db->errmsg, "Error parsing SQL text:\n%s\n", sql);
-    append_oci_error(db->errmsg, db->err);
+    sprintf(db->errmsg, "Error parsing SQL %s:\n---\n%s\n---\n", is_command ? "command" : "query", sql);
+    append_oci_error(db->errmsg, db->err, 0);
+	append_indicator(db, sql);
     return db->success = 0;
   }
 
@@ -651,7 +786,7 @@ static int exec_sql(struct db_conn * db, char * sql, struct sqlparam ** params,
     if ( status != OCI_SUCCESS ) {
       sprintf(db->errmsg, "Error binding SQL input parameter '%s' to value '%s' in query:\n---\n%s\n---\n",
               pname, data, sql);
-      append_oci_error(db->errmsg, db->err);
+      append_oci_error(db->errmsg, db->err, 0);
       return db->success = 0;
     }
 
@@ -667,44 +802,36 @@ static int exec_sql(struct db_conn * db, char * sql, struct sqlparam ** params,
 
   if ( OCI_SUCCESS != status ) {
 
-    ub4 err_offset = 0;
-
     sprintf(db->errmsg, "Error executing SQL %s:\n---\n%s\n---\n", is_command ? "command" : "query", sql);
 
-    /* Also show what params were given */
+    /* Also show what params were given.  Since parameter values may be really big, and our error buffer
+	   is fixed, truncate the parameter values to about 500 bytes (while noting the length if they were
+	   truncated).  This will allow them to be displayed in a reasonably readable way, while at the
+	   same time, showing the length will enable the user to diagnose "value too large for column" problems. */
     if ( nparam ) {
       int i;
-      char buf[50000], buf2[10000];
+      char buf[ERRBUF_BYTES], buf2[10000];
 
       sprintf(buf, "SQL bind parameters (%d) given:\n", nparam);
       strcat(db->errmsg, buf);
       params = params_save;
       for (i=0; i < nparam; i++) {
-        sprintf(buf, "'%s' -> '%s'\n", params[i]->name, params[i]->value.data);
+		char * pdata = params[i]->value.data;
+		int plen = strlen(pdata);
+		if ( plen > ERRBUF_PARAM_TRUNC ) {
+		  sprintf(buf2, "[String of length %d (truncated)] ", plen);
+		  strncat(buf2, pdata, ERRBUF_PARAM_TRUNC);
+		  strcat(buf2, "...");
+		  pdata = buf2;
+		}
+        sprintf(buf, "'%s' -> '%s'\n", params[i]->name, pdata);
         strcat(db->errmsg, buf);
       }
       strcat(db->errmsg, "---\n");
     }
     
-    append_oci_error(db->errmsg, db->err);
-
-    /* Get the parse error offset so we can print an indicator as to
-       where the error occurred */
-    status = OCIAttrGet(db->stmt, OCI_HTYPE_STMT, (dvoid *) &err_offset, (ub4 *) 0, OCI_ATTR_PARSE_ERROR_OFFSET, db->err);
-    if ( OCI_SUCCESS != status ) {
-      strcat(db->errmsg, "Could not get parse error offset:\n");
-      append_oci_error(db->errmsg, db->err);
-    }
-    else {
-      char p[100];
-      sprintf(p, "At character %d, near ==> indicator in:\n----\n", err_offset);
-      strcat(db->errmsg, p);
-
-      strncat(db->errmsg, sql, err_offset);
-      strcat(db->errmsg, "\n==> ");
-      strcat(db->errmsg, sql + err_offset);
-      strcat(db->errmsg, "\n----\n");
-    }
+    append_oci_error(db->errmsg, db->err, 0);
+	append_indicator(db, sql);
     return db->success = 0;
   }
 
@@ -730,7 +857,7 @@ static int exec_sql(struct db_conn * db, char * sql, struct sqlparam ** params,
     attr_status = OCIAttrGet((dvoid *) db->stmt, OCI_HTYPE_STMT, &db->rows_affected, 0, OCI_ATTR_ROW_COUNT, db->err);
     if ( attr_status != OCI_SUCCESS ) {
       sprintf(db->errmsg, "Error getting row count for command:\n");
-      append_oci_error(db->errmsg, db->err);
+      append_oci_error(db->errmsg, db->err, 0);
       return db->success = 0;
     }
   }
@@ -760,7 +887,7 @@ static int init_session(struct db_conn * db, char *user, char * schema, char *pa
   /* If already connected, drop and re-connect */
   if ( db->env )
     disconnect(db);
-
+  
   /* Clear out results */
   *db->errmsg = '\0';
   db->env = 0;
@@ -781,27 +908,39 @@ static int init_session(struct db_conn * db, char *user, char * schema, char *pa
   }
   
   /* Init Oracle library */
-  /* TODO: use mode OCI_OBJECT so that we can describe tables w/ BLOB cols */
+  /* If the Oracle 8 version is really old (< 8.1.5) use these old calls instead of OCIEnvCreate */
+#if 0
   status = OCIInitialize(OCI_OBJECT,
                          0,         /* Context for mode OCI_OBJECT */
                          0,         /* Our malloc() func */
                          0,         /* Our realloc() func */
                          0);        /* Our free() func */
-
-  /* Init an Oracle "environment handle" */
   status = OCIEnvInit(&db->env,     /* Returned: new handle */
                       OCI_DEFAULT,  /* Use mutexes */
                       0,            /* Our memory size */
                       0);           /* Returned: alloc'ed memory */
+#endif
+
+  /* Create the environment */
+  status = OCIEnvCreate(&db->env,     /* Returned: new handle */
+						OCI_OBJECT,   /* Use object features */
+                        0,            /* Context */
+                        0,            /* Our malloc() func */
+                        0,            /* Our realloc() func */
+                        0,            /* Our free() func */
+						0,            /* Our memory size */
+						0);           /* Returned: alloc'ed memory */
   if ( OCI_SUCCESS != status ) {
-    sprintf(db->errmsg, "Error initializing OCI library: OCIEnvInit() returned %s", status);
+    sprintf(db->errmsg, "Error initializing OCI library: OCIEnvCreate/OCIEnvInit() returned %d:\n", status);
+	append_oci_error(db->errmsg, db->env, 1);
     return db->success = 0;
   }
 
   /* Get an error handle */
   status = OCIHandleAlloc(db->env, (dvoid **) &db->err, OCI_HTYPE_ERROR, 0, 0);
   if ( OCI_SUCCESS != status ) {
-    sprintf(db->errmsg, "Error initializing OCI error handle: OCIHandleAlloc(OCI_HTYPE_ERROR) returned %s", status);
+    sprintf(db->errmsg, "Error initializing OCI error handle: OCIHandleAlloc(OCI_HTYPE_ERROR) returned %d:\n", status);
+	append_oci_error(db->errmsg, db->env, 1);
     OCIHandleFree(db->env, OCI_HTYPE_ENV);
     db->env = 0;
     return db->success = 0;
@@ -812,7 +951,7 @@ static int init_session(struct db_conn * db, char *user, char * schema, char *pa
                     user, strlen(user), password, strlen(password), sid, strlen(sid));
   if ( OCI_SUCCESS != status ) {
     sprintf(db->errmsg, "Error logging on to Oracle service '%s' as user '%s':\n", sid, user);
-    append_oci_error(db->errmsg, db->err);
+    append_oci_error(db->errmsg, db->err, 0);
     OCIHandleFree(db->env, OCI_HTYPE_ENV);
     db->env = 0;
     return db->success = 0;
@@ -849,7 +988,7 @@ static int init_session(struct db_conn * db, char *user, char * schema, char *pa
   status = OCIAttrSet(db->stmt, OCI_HTYPE_STMT, &attval, 0, OCI_ATTR_PREFETCH_MEMORY, db->err);
   if ( status != OCI_SUCCESS ) {
     sprintf(db->errmsg, "Error %d setting prefetch memory:\n", status);
-    append_oci_error(db->errmsg, db->err);
+    append_oci_error(db->errmsg, db->err, 0);
     return db->success = 0;
   }
     
@@ -864,7 +1003,7 @@ static int init_session(struct db_conn * db, char *user, char * schema, char *pa
   status = OCIAttrSet(db->stmt, OCI_HTYPE_STMT, &attval, 0, OCI_ATTR_PREFETCH_ROWS, db->err);
   if ( status != OCI_SUCCESS ) {
     sprintf(db->errmsg, "Error settgin prefetch row count:\n");
-    append_oci_error(db->errmsg, db->err);
+    append_oci_error(db->errmsg, db->err, 0);
     return db->success = 0;
   }
 
@@ -883,7 +1022,7 @@ static int get_param_attr(CONST dvoid * param, int ncol, struct db_conn * db, dv
   if ( attr_status != OCI_SUCCESS ) {
     /* Error getting param info */
     sprintf(db->errmsg, "Error getting column %s for col %d:\n", prompt, ncol);
-    append_oci_error(db->errmsg, db->err);
+    append_oci_error(db->errmsg, db->err, 0);
     return db->success = 0;
   }
 
@@ -893,8 +1032,9 @@ static int get_param_attr(CONST dvoid * param, int ncol, struct db_conn * db, dv
 
 /* ------------------------------------------------------------------------------------------------------------- */
 
-/* Append Oracle error to our specific error */
-static void append_oci_error(char *errbuf, OCIError * err)
+/* Append Oracle error to our specific error.  Error pointer is
+   usually (OCIError *) but sometimes (OCIEnv *), if isenv true */
+static void append_oci_error(char *errbuf, void * err, int isenv)
 {
   sword status      = 0;
   char *dummy = 0;
@@ -902,14 +1042,22 @@ static void append_oci_error(char *errbuf, OCIError * err)
   char buf[50000];
 
   /* If Oracle message begins with this, it means we cannot locate
-     error messages.  Probably issue with HOME */
+     error messages.  Probably issue with ORACLE_HOME */
     
   /* Hack: this is highly language dependent */
   char * leading = "Error while trying to retrieve text for error";
+
+  /* Do nothing if don't have a handle */
+  if ( ! err )
+	return;
     
-  status = OCIErrorGet(err, 1, dummy, &errcode, buf, sizeof buf, OCI_HTYPE_ERROR);
-  if ( status )
+  status = OCIErrorGet(err, 1, dummy, &errcode, buf, sizeof buf, isenv ? OCI_HTYPE_ENV : OCI_HTYPE_ERROR);
+  if ( status ) {
+	char reason[1000];
+	sprintf(reason, "[Could not get error message from Oracle: OCIErrorGet() returned %d]\n", status);
+    strcat(errbuf, reason);
     return;
+  }
 
   /* Append Oracle error */
   strcat(errbuf, buf);
@@ -917,20 +1065,47 @@ static void append_oci_error(char *errbuf, OCIError * err)
   /* See if having trouble getting error text */
   if ( 0 == strncmp(buf, leading, strlen(leading)) ) {
     char reason[1000];
-    char * home = getenv("HOME");
+    char * home = getenv("ORACLE_HOME");
     if ( ! home )
-      strcpy(reason, "HOME is not set.");
+      strcpy(reason, "ORACLE_HOME is not set.");
     else if ( ! *home )
-      strcpy(reason, "HOME is set to the empty string.");
+      strcpy(reason, "ORACLE_HOME is set to the empty string.");
     else 
-      sprintf(reason, "HOME value '%s' is possibly not valid.", home);
+      sprintf(reason, "ORACLE_HOME value '%s' is possibly not valid.", home);
 
-    strcat(errbuf, "Cannot get Oracle error message text.  Check HOME environment variable.\n");
+    strcat(errbuf, "Cannot get Oracle error message text.  Check ORACLE_HOME environment variable.\n");
     strcat(errbuf, reason);
   }
 
 }
 
+/* ------------------------------------------------------------------------------------------------------------- */
+/* Append error location indicator to error message */
+void append_indicator(struct db_conn * db, char * sql)
+{
+    ub2		err_offset = 0;
+	sword	status  = 0;
+					  
+    /* Get the parse error offset so we can print an indicator as to
+       where the error occurred */
+    status = OCIAttrGet(db->stmt, OCI_HTYPE_STMT, (dvoid *) &err_offset, (ub4 *) 0, OCI_ATTR_PARSE_ERROR_OFFSET, db->err);
+    if ( OCI_SUCCESS != status ) {
+      strcat(db->errmsg, "Could not get parse error offset:\n");
+      append_oci_error(db->errmsg, db->err, 0);
+    }
+    else {
+      char p[100];
+      sprintf(p, "At character %d, near ==> indicator in:\n----\n", err_offset);
+      strcat(db->errmsg, p);
+
+      strncat(db->errmsg, sql, err_offset);
+      strcat(db->errmsg, "\n==> ");
+      strcat(db->errmsg, sql + err_offset);
+      strcat(db->errmsg, "\n----\n");
+    }
+}
+
+
 /* ------------------------------------------------------------------------------------------------------------- */
 
 /* Decode data type code */
@@ -938,29 +1113,31 @@ static char * decode_data_type(int dtype)
 {
   static char buf[100];
   switch (dtype) {
-  case 1:   return "VARCHAR2";
-  case 2:   return "NUMBER";
-  case 3:   return "INTEGER";
-  case 4:   return "FLOAT";
-  case 8:   return "LONG";
-  case 11:  return "ROWID";
-  case 12:  return "DATE";
-  case 23:  return "RAW";
-  case 24:  return "LONG RAW";
-  case 96:  return "CHAR";
-  case 104: return "ROWID DESC";
-  case 105: return "MLSLABEL";
-  case 108: return "User defined";
-  case 111: return "REF";
-  case 112: return "CLOB";
-  case 113: return "BLOB";
+  case 1:   return "VARCHAR2";             /* SQLT_CHR */
+  case 2:   return "NUMBER";               /* SQLT_NUM */
+  case 3:   return "INTEGER";              /* SQLT_INT */
+  case 4:   return "FLOAT";                /* SQLT_FLT */
+  case 8:   return "LONG";                 /* SQLT_LNG */
+  case 11:  return "ROWID";                /* SQLT_RID -- through Oracle 7 */
+  case 12:  return "DATE";                 /* SQLT_DAT */
+  case 23:  return "RAW";                  /* SQLT_BIN */
+  case 24:  return "LONG RAW";             /* SQLT_LBI */
+  case 96:  return "CHAR";                 /* SQLT_AFC (ANSI Fixed Char) */
+  case 104: return "ROWID DESC";           /* SQLT_RDD (RowId Descriptor) */
+  case 105: return "MLSLABEL";             /* SQLT_LAB */
+  case 108: return "User defined";         /* SQLT_NTY (Named Type) */
+  case 111: return "REF";                  /* ? */
+  case 112: return "CLOB";                 /* SQLT_CLOB - Character LOB */
+  case 113: return "BLOB";                 /* SQLT_BLOB - Binary LOB */
+  case 114: return "BFILE";                /* SQLT_BFILEE - Binary file LOB */
+  case 115: return "CFILE";                /* SQLT_CFILEE - Character file LOB */
   }
   sprintf("<Unknown type %d>", buf);
   return buf;
 }
 
 /* Get appropriate length for fetch into on output define variable */
-static int fetch_data_len(int dtype, int dlen)
+static int fetch_data_len(int dtype, int dlen, int long_len)
 {
   static char buf[100];
 
@@ -968,29 +1145,56 @@ static int fetch_data_len(int dtype, int dlen)
      the largest buffer that would ever be used by an Oracle number,
      for dates a reasonable limit for formatted dates (e.g.,
      "Wednesday, November 23, 2005") and for chars, just that length,
-     plus one for the null terminator.
+     plus one for the null terminator.  For LONG, use configured length,
+	 after which Oracle will truncate.  For RAW, use double length so
+	 hex conversion will fit.
 
      With all this, we should not be getting Oracle truncation due to
      deficiencies in our buffer size (i.e., our buffer should always
-     accomodate the Oracle data).  */
+     accomodate the Oracle data), with the exception of LONG. */
 
   switch (dtype) {
-  case 1:   /* VARCHAR2 */
-  case 96:  /* CHAR */
+  case 1:   /* VARCHAR2 (SQLT_CHR) */
+  case 96:  /* CHAR (SQLT_AFC) */
     return dlen + 1;
     
-  case 2:   /* NUMBER */
-  case 3:   /* INTEGER */
-  case 4:   /* FLOAT */
-    return 50;
+	/* Use 1000 for numbers so that very large/small floats will never be truncated */
+  case 2:   /* NUMBER (SQLT_NUM) */
+  case 3:   /* INTEGER (SQLT_INT) */
+  case 4:   /* FLOAT (SQLT_FLT) */
+    return 1000;
     
-  case 8:   /* LONG */
-    return 70000;
+  case 104: /* ROWID DESC (SQLT_RDD) */
+	return 50;
+
+  case 8:   /* LONG (SQLT_LNG) */
+  case 112: /* CLOB (SQLT_CLOB) */
+    return long_len + 1;
     
-  case 12:  /* DATE */
+	/* Use twice long_len for BLOBs so we can convert to hex */
+  case 113: /* BLOB (SQLT_BLOB) */
+  case 114: /* BFILE (SQLT_FILE) */
+    return 2 * long_len + 1;
+
+  case 24:  /* LONG RAW (SQLT_LBI) */
+    return 2 * long_len + 1;
+
+  case 23:  /* RAW (returned in hex) (SQLT_BIN) */
+	return 2 * dlen + 1;
+    
+  case 12:  /* DATE (SQLT_DAT) */
     return 50;
   }
-  return SQLT_STR;
+  return 0;
+}
+
+/* Is data type a blob type? */
+static int is_blob_type(int dtype)
+{
+  return    dtype == SQLT_CLOB
+	     || dtype == SQLT_BLOB
+	     || dtype == SQLT_BFILE
+	;
 }
 
 /* ------------------------------------------------------------------------------------------------------------- */
@@ -1000,6 +1204,9 @@ static void free_column(struct column * c)
   free_if_non_null(c->name);
   free_if_non_null(c->data);
 
+  if ( c->lob_locator )
+	OCIDescriptorFree(c->lob_locator, OCI_DTYPE_LOB);
+
   if ( c->def )
     OCIHandleFree(c->def, OCI_HTYPE_DEFINE);
 }
@@ -1228,6 +1435,24 @@ static int empty(char * s) { return !s || !*s; }
 /* Map NULL to empty string */
 static char * valid_string(char * s) { return s ? s : ""; }
 
+/* Convert a binary buffer to null-terminated hex in place.  Actual
+   allocation of buf must be twice N plus 1.  Work from the end to the
+   start so we don't stomp on ourselves. */
+/* Convert int to hex */
+#define itoh(n)(((n) > 9) ? ((n) - 10 + 'A') : ((n) + '0'))
+static void hexify (unsigned char * buf, int n) 
+{
+  int i;
+  for ( i = n-1; i >= 0; i-- ) {
+	unsigned char c = buf[i];
+    buf[2*i] = itoh(c / 16);
+    buf[2*i + 1] = itoh(c & 0xF);
+  }
+
+  /* Null terminate */
+  buf[2*n] = '\0';
+}
+
 /* ------------------------------------------------------------------------------------------------------------- */
 
 /* Wrapper routines - these are part of the "void *" based interface
@@ -1236,8 +1461,8 @@ static char * valid_string(char * s) { return s ? s : ""; }
    implemented in the interface module.
 */
 
-void * oracle_connect(char * user, char * schema, char * password, char * sid, int prefetch_bytes, int auto_commit)
-{ return (void *) connect(user, schema, password, sid, prefetch_bytes, auto_commit); }
+void * oracle_connect(char * user, char * schema, char * password, char * sid, int prefetch_bytes, int auto_commit, int long_len, int truncate_ok)
+{ return (void *) connect(user, schema, password, sid, prefetch_bytes, auto_commit, long_len, truncate_ok); }
 
 int oracle_disconnect(void * db)
 { return disconnect((struct db_conn *) db); }
