@@ -314,7 +314,7 @@
   (when (or (eq situation 't) direct-superclasses-p)
     ; Check the direct-superclasses.
     (unless (proper-list-p direct-superclasses)
-      (error (TEXT "(~S ~S) for class ~S: The ~S argument should be a proper list, not ~S")
+      (error #1=(TEXT "(~S ~S) for class ~S: The ~S argument should be a proper list, not ~S")
              (if (eq situation 't) 'initialize-instance 'shared-initialize)
              'class name ':direct-superclasses direct-superclasses))
     (unless (every #'(lambda (x)
@@ -337,8 +337,7 @@
   (when (or (eq situation 't) direct-slots-as-lists-p)
     ; Check the direct-slots.
     (unless (proper-list-p direct-slots-as-lists)
-      (error (TEXT "(~S ~S) for class ~S: The ~S argument should be a proper list, not ~S")
-             (if (eq situation 't) 'initialize-instance 'shared-initialize)
+      (error #1# (if (eq situation 't) 'initialize-instance 'shared-initialize)
              'class name ':direct-slots direct-slots-as-lists))
     (dolist (sl direct-slots-as-lists)
       (unless (canonicalized-slot-p sl)
@@ -348,8 +347,7 @@
   (when (or (eq situation 't) direct-default-initargs-p)
     ; Check the direct-default-initargs.
     (unless (proper-list-p direct-default-initargs)
-      (error (TEXT "(~S ~S) for class ~S: The ~S argument should be a proper list, not ~S")
-             (if (eq situation 't) 'initialize-instance 'shared-initialize)
+      (error #1# (if (eq situation 't) 'initialize-instance 'shared-initialize)
              'class name ':direct-default-initargs direct-default-initargs))
     (dolist (definitarg direct-default-initargs)
       (unless (canonicalized-default-initarg-p definitarg)
