@@ -11,10 +11,10 @@
 (use-package '("COMMON-LISP" "EXT") "FFI")
 (in-package "FFI")
 
-(export '(def-c-type def-c-var parse-c-type deparse-c-type
+(export '(def-c-type def-c-var def-c-const parse-c-type deparse-c-type
           def-c-call-out def-call-out #+AFFI def-lib-call-out
           def-c-call-in def-call-in default-foreign-language
-          c-lines *output-c-functions* *output-c-variables*
+          c-lines *output-c-functions* *output-c-variables* *foreign-guard*
           nil boolean character char uchar short ushort int uint long ulong
           uint8 sint8 uint16 sint16 uint32 sint32 uint64 sint64
           single-float double-float
@@ -528,6 +528,10 @@
   (init-once '())
   (init-always '())
   (fini '())
+  ;; type -> (function-name . #(const1 const2 const3 ...))
+  (constant-table (make-hash-table :test 'stablehash-eq
+                                   :key-type 'symbol :value-type 'string
+                                   :warn-if-needs-rehash-after-gc t))
   (variable-list '())
   (function-list '()))
 (define-symbol-macro *name*
@@ -544,6 +548,8 @@
     (ffi-module-init-always *ffi-module*))
 (define-symbol-macro *fini*
     (ffi-module-fini *ffi-module*))
+(define-symbol-macro *constant-table*
+    (ffi-module-constant-table *ffi-module*))
 (define-symbol-macro *variable-list*
     (ffi-module-variable-list *ffi-module*))
 (define-symbol-macro *function-list*
@@ -567,8 +573,9 @@
                       (symbol-package name)
                       (package-case-inverted-p (symbol-package name)))
                (cs-cl:string name)
-               (string name)))
-  (unless (some #'lower-case-p name) (setq name (string-downcase name)))
+               (let ((nm (string name)))
+                 (if (some #'lower-case-p nm)
+                     nm (string-downcase name)))))
   (if (c-ident-p name)
     name
     (with-output-to-string (s)
@@ -599,13 +606,12 @@
 (defun to-c-typedecl (c-type name)
   (case c-type
     ((nil) (format nil "void ~A" name))
-    (boolean (format nil "int ~A" name))
+    ((boolean int) (format nil "int ~A" name))
     (character (format nil "char ~A" name))
     ((char sint8) (format nil "sint8 ~A" name))
     ((uchar uint8) (format nil "uint8 ~A" name))
     ((short sint16) (format nil "sint16 ~A" name))
     ((ushort uint16) (format nil "uint16 ~A" name))
-    (int (format nil "int ~A" name))
     (uint (format nil "unsigned int ~A" name))
     (long (format nil "long ~A" name))
     (ulong (format nil "unsigned long ~A" name))
@@ -668,6 +674,7 @@
 
 (defvar *output-c-functions* nil)
 (defvar *output-c-variables* nil)
+(defvar *foreign-guard* nil)
 
 (defun prepare-module ()
   (unless *ffi-module*
@@ -703,6 +710,21 @@
       (format *coutput-stream* "uintC module__~A__object_tab_size = ~D;~%"
               *c-name* count))
     (format *coutput-stream* "~%")
+    (maphash (lambda (type fun-vec)
+               (let* ((fun (car fun-vec)) (vec (cdr fun-vec))
+                      (c-decl (to-c-typedecl type fun)))
+                 (format *coutput-stream* "~A (int number, int *definedp);~%~
+                                           ~A (int number, int *definedp) {
+  *definedp=1;~%  switch (number) {~%"
+                         c-decl c-decl)
+                 (dotimes (num (length vec))
+                   (let ((const (aref vec num)))
+                     (format *coutput-stream* "#  if defined(~A)
+    case ~D: return ~A;~%#  endif~%"
+                             const num const)))
+                 (format *coutput-stream* "    default: *definedp=0; return 0;
+  }~%}~%#define HAVE_~A~%" (string-upcase (symbol-name fun)))))
+             *constant-table*)
     (setq *variable-list*
           (nreverse (delete-duplicates
                      *variable-list* :key #'first :test #'equal)))
@@ -741,15 +763,23 @@
             *c-name* *c-name* *c-name*
             *c-name* *init-once* *c-name* *init-always*)
     (dolist (variable *variable-list*)
-      (format *coutput-stream*
-              "  register_foreign_variable((void*)&~A,~A,~D,sizeof(~A));~%"
-              (first variable) (to-c-string (first variable))
-              (third variable) (first variable)))
+      (let ((c-name (first variable)))
+        (when *foreign-guard*
+          (format *coutput-stream* "# if HAVE_DECL_~A~%"
+                  (string-upcase c-name)))
+        (format *coutput-stream*
+                "  register_foreign_variable((void*)&~A,~A,~D,sizeof(~A));~%"
+                c-name (to-c-string c-name) (third variable) (first variable))
+        (when *foreign-guard* (format *coutput-stream* "# endif~%"))))
     (dolist (function *function-list*)
-      (format *coutput-stream*
-              "  register_foreign_function((void*)&~A,~A,~D);~%"
-              (first function) (to-c-string (first function))
-              (svref (second function) 3)))
+      (let ((c-name (first function)))
+        (when *foreign-guard*
+          (format *coutput-stream* "# if defined(HAVE_~A)~%"
+                  (string-upcase c-name)))
+        (format *coutput-stream*
+                "  register_foreign_function((void*)&~A,~A,~D);~%"
+                c-name (to-c-string c-name) (svref (second function) 3))
+        (when *foreign-guard* (format *coutput-stream* "# endif~%"))))
     (format *coutput-stream*
             "}~2%void module__~A__fini_function (module_t* module)~%~
             {~{~%~A~}~%}~%"
@@ -800,6 +830,52 @@
     (parse-foreign-name (second name-option))
     (to-c-name lisp-name)))
 
+;; CPP consts (#define'd in an *.h file):
+;; for each type (int, string, pointer) there is a C function (and a
+;;  foreign-function created when the first constant of this type is
+;;  encountered) mapping a number (assigned at _compile_ time) into the value
+;; when loading a compiled file, the C function has already been writted
+;;  into the C file and compiled, so the numbers have to be pre-assigned
+(defmacro DEF-C-CONST (&whole whole-form name &rest options)
+  (setq name (check-symbol name (first whole-form)))
+  (prepare-module)
+  (let* ((alist (parse-options options '(:name :type :documentation)
+                               whole-form))
+         (doc (cdr (assoc ':documentation alist))) ; ("doc string") or NIL
+         (c-type (or (second (assoc ':type alist)) 'ffi:int))
+         (c-name (foreign-name name (assoc ':name alist)))
+         f-name c-number)
+    (check-type c-type (member ffi:int ffi:c-string ffi:c-pointer)
+                "A constant must be either an integer, a string or a pointer")
+    (setq f-name (intern
+                  (format nil "module__~A__constant_map_~A" *name*
+                          (nstring-downcase
+                           (nsubstitute #\_ #\-
+                                        (copy-seq (symbol-name c-type))))))
+          c-number
+          (vector-push-extend
+           c-name (cdr (or (gethash c-type *constant-table*)
+                           (setf (gethash c-type *constant-table*)
+                                 (cons f-name (make-array 10 :adjustable t
+                                                          :fill-pointer 0)))))))
+    `(progn
+       ;; c-number == 0 ==> need to output the def-call-out form
+       ,@(when (zerop c-number)
+           `((ffi:def-call-out ,f-name
+                 (:arguments (number ffi:int)
+                             (defined-p (ffi:c-ptr ffi:int) :out))
+               (:return-type ,c-type))))
+       (defconstant ,name (c-const-value ',f-name ,c-number ',name ',c-name)
+         ,@doc))))
+
+(defun c-const-value (f-name c-number name c-name) ; ABI
+  (multiple-value-bind (value value-p) (funcall f-name c-number)
+    (if value-p value
+        (progn
+          (warn (TEXT "~S(~S): CPP constant ~A is not defined")
+                'def-c-const name c-name)
+          (sys::%unbound)))))
+
 (defmacro DEF-C-VAR (&whole whole-form
                      name &rest options)
   (setq name (check-symbol name (first whole-form)))
@@ -827,10 +903,14 @@
          (getter-function-name (sys::symbol-suffix name "%GETTER%"))
          (setter-function-name (sys::symbol-suffix name "%SETTER%"))
          |#
-        )
-    `(PROGN
-       ,(unless library
-          `(EVAL-WHEN (COMPILE) (NOTE-C-VAR ',c-name ',type ',flags)))
+         (def (gensym "DEF-C-VAR-")))
+    `(LET ((,def (LOAD-TIME-VALUE
+                  ,(if library
+                       `(FFI::FOREIGN-LIBRARY-VARIABLE
+                         ',c-name (FFI::FOREIGN-LIBRARY ,library)
+                         nil (PARSE-C-TYPE ',type))
+                       `(LOOKUP-FOREIGN-VARIABLE
+                         ',c-name (PARSE-C-TYPE ',type))))))
        #|
        (LET ((FVAR (LOOKUP-FOREIGN-VARIABLE ',c-name (PARSE-C-TYPE ',type))))
          (DEFUN ,getter-function-name () (FOREIGN-VALUE FVAR))
@@ -841,16 +921,13 @@
        (DEFSETF ,getter-function-name ,setter-function-name)
        (DEFINE-SYMBOL-MACRO ,name (,getter-function-name))
        |#
-       (SYSTEM::%PUT ',name 'FOREIGN-VARIABLE
-         (LOAD-TIME-VALUE
-           ,(if library
-              `(FFI::FOREIGN-LIBRARY-VARIABLE
-                ',c-name (FFI::FOREIGN-LIBRARY ,library)
-                nil (PARSE-C-TYPE ',type))
-              `(LOOKUP-FOREIGN-VARIABLE ',c-name (PARSE-C-TYPE ',type)))))
-       ,@(when doc `((SETF (DOCUMENTATION ',name 'VARIABLE) ',(second doc))))
-       (DEFINE-SYMBOL-MACRO ,name
-         (FOREIGN-VALUE (LOAD-TIME-VALUE (GET ',name 'FOREIGN-VARIABLE))))
+       ,(unless library
+          `(EVAL-WHEN (COMPILE) (NOTE-C-VAR ',c-name ',type ',flags)))
+       (when ,def
+         (SYSTEM::%PUT ',name 'FOREIGN-VARIABLE ,def)
+         ,@(when doc `((SETF (DOCUMENTATION ',name 'VARIABLE) ',(second doc))))
+         (DEFINE-SYMBOL-MACRO ,name
+             (FOREIGN-VALUE (LOAD-TIME-VALUE (GET ',name 'FOREIGN-VARIABLE)))))
        ',name)))
 
 (defun note-c-var (c-name type flags) ; ABI
@@ -945,8 +1022,9 @@
   (setq name (check-symbol name (first whole-form)))
   (let* ((alist
           (parse-options options '(:name :arguments :return-type :language
-                                    :built-in :library :documentation)
-                          whole-form))
+                                   :built-in :library :documentation)
+                         whole-form))
+         (def (gensym "DEF-CALL-OUT-"))
          (doc (assoc ':documentation alist))
          (parsed-function (parse-c-function alist whole-form))
          (signature (argvector-to-signature (svref parsed-function 2)))
@@ -954,19 +1032,18 @@
          (c-name (foreign-name name (assoc :name alist))))
     (setq alist (remove-if (lambda (el) (sys::memq (car el) '(:name :library)))
                            alist))
-    `(PROGN
+    `(LET ((,def ,(if library
+                      `(FFI::FOREIGN-LIBRARY-FUNCTION
+                        ',c-name (FFI::FOREIGN-LIBRARY ,library)
+                        nil ,parsed-function)
+                      `(LOOKUP-FOREIGN-FUNCTION ',c-name ,parsed-function))))
        ,(unless library
           `(EVAL-WHEN (COMPILE) (NOTE-C-FUN ',c-name ',alist ',whole-form)))
-       (LET ()
+       (when ,def
          (SYSTEM::REMOVE-OLD-DEFINITIONS ',name)
          (COMPILER::EVAL-WHEN-COMPILE (COMPILER::C-DEFUN ',name ',signature))
          ,@(when doc `((SETF (DOCUMENTATION ',name 'FUNCTION) ',(second doc))))
-         (SYSTEM::%PUTD ',name
-            ,(if library
-               `(FFI::FOREIGN-LIBRARY-FUNCTION
-                 ',c-name (FFI::FOREIGN-LIBRARY ,library)
-                 nil ,parsed-function)
-               `(LOOKUP-FOREIGN-FUNCTION ',c-name ,parsed-function))))
+         (SYSTEM::%PUTD ',name ,def))
        ',name)))
 
 (defun note-c-fun (c-name alist whole) ; ABI
@@ -1014,6 +1091,11 @@
          (NOTE-C-CALL-IN ',name ',c-name ',alist ',whole-form))
        ',name)))
 
+(defun convert-to-foreign-C (flags)
+  (if (flag-set-p flags ff-flag-malloc-free)
+      "convert_to_foreign_mallocing"
+      "convert_to_foreign_nomalloc"))
+
 (defun note-c-call-in (name c-name alist whole) ; ABI
   (when (compiler::prepare-coutput-file)
     (prepare-module)
@@ -1064,9 +1146,7 @@
           (format *coutput-stream* " {~%  var ~A;~%~:
   ~A(~A,value1,&retval);~%"
                   (to-c-typedecl rettype "retval")
-                  (if (flag-set-p flags ff-flag-malloc-free)
-                    "convert_to_foreign_mallocing"
-                    "convert_to_foreign_nomalloc")
+                  (convert-to-foreign-C flags)
                   (object-to-c-value (pass-object rettype))))
         (let ((outargcount (if (eq rettype 'NIL) 0 1)))
           (mapc #'(lambda (argtype argflag argname)
@@ -1078,9 +1158,7 @@
                               (if (eql outargcount 0) ""
                                 (format nil "if (mv_count >= ~D) "
                                         (+ outargcount 1)))
-                              (if (flag-set-p argflag ff-flag-malloc-free)
-                                "convert_to_foreign_mallocing"
-                                "convert_to_foreign_nomalloc")
+                              (convert-to-foreign-C argflag)
                               (object-to-c-value
                                 (pass-object (svref argtype 1)))
                               (if (eql outargcount 0)
