@@ -13,13 +13,14 @@
   (do-all-symbols (sym) (remprop sym 'sys::definition))
   (when (fboundp 'clos::install-dispatch)
     (do-all-symbols (sym)
-      (when (and (fboundp sym) (clos::generic-function-p (symbol-function sym)))
+      (when (and (fboundp sym)
+                 (typep (symbol-function sym) clos::<standard-generic-function>))
         (let ((gf (symbol-function sym)))
           (when (clos::gf-never-called-p gf)
             (clos::install-dispatch gf)
   ) ) ) ) )
   (setq - nil + nil ++ nil +++ nil * nil ** nil *** nil / nil // nil /// nil)
-  (savemem "lispimag.mem")
+  (savemem "lispimag.mem" nil)
   (room nil)
 )
 
@@ -27,10 +28,17 @@
 ;; This function works only when compiled!
 (defun saveinitmem (&optional (filename "lispinit.mem")
                     &key ((:quiet *quiet*) nil) init-function verbose
-                    ((:start-package *package*) *package*)
-                    (locked-packages *system-package-list*))
-  (let* ((old-driver *driver*)
-         (fn (merge-pathnames filename #.(make-pathname :type "mem")))
+                    ((:norc *norc*) nil)
+                    keep-global-handlers (start-package *package*)
+                    (locked-packages *system-package-list*) executable)
+  (let* ((old-driver *driver*) old-global-handlers
+         (*package* (sys::%find-package start-package))
+         (fn (if (not executable)
+               (merge-pathnames filename #.(make-pathname :type "mem"))
+               ;; win32 executables require "exe" extension
+               #+(or win32 cygwin)
+               (make-pathname :type "exe" :defaults filename)
+               #-(or win32 cygwin) filename))
          (*driver*
            #'(lambda ()
                ;(declare (special *command-index* *home-package*
@@ -58,11 +66,15 @@
                (setq *driver* old-driver)
                (when init-function (funcall init-function))
                (funcall *driver*))))
+    (unless keep-global-handlers
+      (setq old-global-handlers ; disable and save all global handlers
+            (ext::set-global-handler nil nil)))
     (setf (package-lock locked-packages) t)
-    (savemem fn)
+    (savemem fn executable)
+    (when old-global-handlers   ; re-install all global handlers
+      (ext::set-global-handler old-global-handlers nil))
     (when verbose
       (fresh-line)
-      (write-string (TEXT "Wrote the memory image into "))
-      (princ fn)
+      (format t (TEXT "Wrote the memory image into ~A") fn)
       (terpri)))
   (room nil))
